"""Wrapper around executing external commands."""

from __future__ import annotations

import subprocess
from datetime import timedelta
from pathlib import Path
from shlex import join

import structlog

from ..constants import APP_NAME
from ..exceptions import CommandFailedError, CommandTimedOutError

__all__ = ["Command"]


class Command:
    """Wrapper around executing external commands.

    This class provides a generic wrapper around subprocess that is
    helpful for executing external commands, checking their status,
    and optionally capturing their output in a consistent way.

    It represents a slight simplification of the standard ~subprocess.run()
    function. It is intended for use by storage classes that perform
    operations via external commands, which are expected to constrain the
    arguments to the command and the parameters to the run() call.
    """

    def __init__(
        self,
        *,
        capture_output: bool = True,
        ignore_fail: bool = False,
        text: bool = True,
        logger: structlog.BoundLogger | None = None,
    ) -> None:
        """Class to execute commands on behalf of a caller.

        Parameters we do not expect to vary over the caller's lifetime
        are set here, while those that are likely to change
        execution-by-execution are given as arguments to run().

        Parameters
        ----------
        capture_output
            If `True`, collect the process's standard output and standard
            error in the returned object.  If `False`, these streams will
            be inherited from the caller.
        ignore_fail
            If `True`, do not raise an exception on command failure.  If
            the process doesn't run at all, ~subprocess.SubprocessError will
            still be raised.
        text
            If `True`, captured stdout and stderr will be a string rather than
            a bytes sequence.
        """
        if logger is None:
            self._logger = structlog.get_logger(APP_NAME)
        else:
            self._logger = logger
        self._capture_output = capture_output
        self._ignore_fail = ignore_fail
        self._text = text

    def run(
        self,
        *args: str,
        cwd: Path | None = None,
        env: dict[str, str] | None = None,
        timeout: timedelta | float | None = None,
    ) -> subprocess.CompletedProcess:
        """Run the command with the provided arguments.

        Parameters
        ----------
        *args
            Arguments to the command.  The first argument is the command
            itself, and if not fully-qualified must exist somewhere in the
            caller's `PATH`.
        cwd
            If provided, change working directories to this path before
            running the command.
        env
            If provided, use this dictionary as the process environment.  If
            not, use the caller's process environment.
        timeout
            If given, the command will be terminated and a
            `~lsst.rsp.startup.exceptions.CommandTimedOutError` will be
            raised if execution time exceeds this timeout.

        Raises
        ------
        CommandFailedError
            Raised if the command failed and ``ignore_fail`` was not set to
            `True`.
        CommandTimedOutError
            Raised if ``timeout`` was given and the command took longer than
            that to complete.
        subprocess.SubprocessError
            Raised if the command could not be executed at all.

        """
        check = not self._ignore_fail
        if isinstance(timeout, timedelta):
            timeout = timeout.total_seconds()
        try:
            self._logger.debug(
                f"Running '{join(args)}'",
                cwd=cwd,
                env=env,
                check=check,
                timeout=timeout,
                capture_output=self._capture_output,
                text=self._text,
            )
            result = subprocess.run(
                args,
                cwd=cwd,
                env=env,
                check=check,
                timeout=timeout,
                capture_output=self._capture_output,
                text=self._text,
            )

        except subprocess.CalledProcessError as e:
            raise CommandFailedError(args, e) from e
        except subprocess.TimeoutExpired as e:
            raise CommandTimedOutError(args, e) from e
        if result.returncode != 0:
            self._logger.error(f"Command '{join(args)}' failed", proc=result)
        else:
            self._logger.debug(
                f"Command '{join(args)}' succeeded", proc=result
            )
        return result



"""Exceptions for the LabRunner startup service."""

from __future__ import annotations

import subprocess
from collections.abc import Iterable
from shlex import join

__all__ = [
    "CommandFailedError",
    "CommandTimedOutError",
]


class CommandFailedError(Exception):
    """Execution of a command failed.

    Parameters
    ----------
    args
        Command (args[0]) and arguments to that command.
    exc
        Exception reporting the failure.

    Attributes
    ----------
    stdout
        Standard output from the failed command.
    stderr
        Standard error from the failed command.
    """

    def __init__(
        self,
        args: Iterable[str],
        exc: subprocess.CalledProcessError,
    ) -> None:
        args_str = join(args)
        msg = f"'{args_str}' failed with status {exc.returncode}"
        super().__init__(msg)
        self.stdout = exc.stdout
        self.stderr = exc.stderr


class CommandTimedOutError(Exception):
    """Execution of a command failed.

    Parameters
    ----------
    args
        Command (args[0]) and arguments to that command.
    exc
        Exception reporting the failure.

    Attributes
    ----------
    stdout
        Standard output from the failed command.
    stderr
        Standard error from the failed command.
    """

    def __init__(
        self,
        args: Iterable[str],
        exc: subprocess.TimeoutExpired,
    ) -> None:
        args_str = join(args)
        msg = f"'{args_str}' timed out after {exc.timeout}s"
        super().__init__(msg)
        self.stdout = exc.stdout
        self.stderr = exc.stderr



"""Helpers for RSP startup logging; this is a stripped-down version of
what Safir does.
"""

import logging
import sys
from typing import Any

import structlog
from structlog.types import EventDict

from ..constants import APP_NAME

__all__ = ["configure_logging"]


def configure_logging(*, debug: bool = False) -> None:
    """Stripped-down version of Safir's "configure_logging()"; we
    always add timestamps, and if we're not running with debugging,
    it's INFO-level JSON logs; if we are running with debugging,
    DEBUG-level text-format.

    Parameters
    ----------
    debug
        Enable debugging?  See above for effect.
    """
    log_level = "DEBUG" if debug else "INFO"
    stream_handler = logging.StreamHandler(stream=sys.stdout)
    stream_handler.setFormatter(logging.Formatter("%(message)s"))
    logger = logging.getLogger(APP_NAME)
    logger.handlers = []
    logger.addHandler(stream_handler)
    logger.setLevel(log_level)
    processors: list[Any] = [
        structlog.stdlib.filter_by_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.PositionalArgumentsFormatter(),
    ]
    processors.append(structlog.processors.TimeStamper(fmt="iso"))
    processors.extend(
        [
            structlog.processors.StackInfoRenderer(),
            structlog.processors.UnicodeDecoder(),
        ]
    )
    if debug:
        # Key-value formatted logging
        processors.append(structlog.stdlib.add_log_level)
        processors.append(structlog.dev.ConsoleRenderer())
    else:
        # JSON-formatted logging
        processors.append(add_log_severity)
        processors.append(structlog.processors.format_exc_info)
        processors.append(structlog.processors.JSONRenderer())
    structlog.configure(
        processors=processors,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


def add_log_severity(
    logger: logging.Logger, method_name: str, event_dict: EventDict
) -> EventDict:
    """Add the log level to the event dict as ``severity``.

    Intended for use as a structlog processor.

    This is the same as `structlog.stdlib.add_log_level` except that it
    uses the ``severity`` key rather than ``level`` for compatibility with
    Google Log Explorer and its automatic processing of structured logs.

    Parameters
    ----------
    logger
        The wrapped logger object.
    method_name
        The name of the wrapped method (``warning`` or ``error``, for
        example).
    event_dict
        Current context and current event. This parameter is also modified in
        place, matching the normal behavior of structlog processors.

    Returns
    -------
    ``structlog.types.EventDict``
        The modified ``structlog.types.EventDict`` with the added key.

    Notes
    -----
    This is stolen directly from Safir.
    """
    severity = structlog.stdlib.add_log_level(logger, method_name, {})["level"]
    event_dict["severity"] = severity
    return event_dict



"""Run a command noninteractively from a configuration JSON file."""

import json
import os
import sys
from pathlib import Path
from typing import Self


class NonInteractiveExecutor:
    """Launch noninteractive Lab container execution from a
    configuration document.
    """

    def __init__(
        self,
        kernel: str,
        command: list[str],
    ) -> None:
        self._kernel = kernel
        self._command = command

    @classmethod
    def from_config(cls, config: Path) -> Self:
        """Load configuration from a JSON document."""
        with config.open() as f:
            obj = json.load(f)
            if obj["type"] != "command":
                raise NotImplementedError(
                    "Only 'command' type noninteractive execution is supported"
                )
            return cls(kernel=obj["kernel"], command=obj["command"])

    def execute(self, env: dict[str, str] | None = None) -> None:
        """Run the command specified in the object, with a supplied
        environment (defaulting to the ambient environment).
        """
        if env is None:
            env = dict(os.environ)
        sys.stdout.flush()
        sys.stderr.flush()
        os.execve(self._command[0], self._command, env=env)



"""Utility functions for LSST JupyterLab notebook environment."""

import os
from contextlib import suppress
from pathlib import Path
from typing import Any
from urllib.parse import urljoin

import pyvo
import requests
from deprecated import deprecated


def format_bytes(n: int) -> str:
    """Format bytes as text.

    Taken from ``dask.distributed``, where it is not exported.

    Examples
    --------
    >>> format_bytes(1)
    '1 B'
    >>> format_bytes(1234)
    '1.23 kB'
    >>> format_bytes(12345678)
    '12.35 MB'
    >>> format_bytes(1234567890)
    '1.23 GB'
    >>> format_bytes(1234567890000)
    '1.23 TB'
    >>> format_bytes(1234567890000000)
    '1.23 PB'
    """
    if n > 1e15:
        return "%0.2f PB" % (n / 1e15)
    if n > 1e12:
        return "%0.2f TB" % (n / 1e12)
    if n > 1e9:
        return "%0.2f GB" % (n / 1e9)
    if n > 1e6:
        return "%0.2f MB" % (n / 1e6)
    if n > 1e3:
        return "%0.2f kB" % (n / 1000)
    return "%d B" % n


def get_hostname() -> str:
    """Return hostname or, failing that, ``localhost``."""
    return os.environ.get("HOSTNAME") or "localhost"


def get_service_url(name: str, env_name: str | None = None) -> str:
    """Get our best guess at the URL for the requested service."""
    if not env_name:
        env_name = name.upper()

    url = os.getenv(f"EXTERNAL_{env_name}_URL")
    if url:
        return url

    base = os.getenv("EXTERNAL_INSTANCE_URL") or ""
    path = os.getenv(f"{env_name}_ROUTE") or f"api/{name}"
    return urljoin(base, path)


def get_pyvo_auth() -> pyvo.auth.authsession.AuthSession | None:
    """Create a PyVO-compatible auth object."""
    tap_url = get_service_url("tap")
    obstap_url = get_service_url("obstap")
    ssotap_url = get_service_url("ssotap")
    siav2_url = get_service_url("siav2")
    s = requests.Session()
    tok = get_access_token()
    if not tok:
        return None
    s.headers["Authorization"] = "Bearer " + tok
    auth = pyvo.auth.authsession.AuthSession()
    auth.credentials.set("lsst-token", s)
    auth.add_security_method_for_url(get_service_url("cutout"), "lsst-token")
    auth.add_security_method_for_url(get_service_url("datalink"), "lsst-token")
    auth.add_security_method_for_url(siav2_url, "lsst-token")
    auth.add_security_method_for_url(siav2_url + "/query", "lsst-token")
    auth.add_security_method_for_url(tap_url, "lsst-token")
    auth.add_security_method_for_url(tap_url + "/sync", "lsst-token")
    auth.add_security_method_for_url(tap_url + "/async", "lsst-token")
    auth.add_security_method_for_url(tap_url + "/tables", "lsst-token")
    auth.add_security_method_for_url(obstap_url, "lsst-token")
    auth.add_security_method_for_url(obstap_url + "/sync", "lsst-token")
    auth.add_security_method_for_url(obstap_url + "/async", "lsst-token")
    auth.add_security_method_for_url(obstap_url + "/tables", "lsst-token")
    auth.add_security_method_for_url(ssotap_url, "lsst-token")
    auth.add_security_method_for_url(ssotap_url + "/sync", "lsst-token")
    auth.add_security_method_for_url(ssotap_url + "/async", "lsst-token")
    auth.add_security_method_for_url(ssotap_url + "/tables", "lsst-token")
    return auth


@deprecated(
    reason="get_pod() always returns None in RSPs running Nubladov3 or later"
)
def get_pod() -> None:
    """Get the name of the running pod (deprecated, not functional).

    No longer useful.  Formerly used to return the Kubernetes object for the
    pod in which this code was running.
    """
    return


def get_node() -> str:
    """Return the name of the current Kubernetes node.

    Returns
    -------
    str
        Name of the Kubernetes node on which this code is running, or the
        empty string if the node could not be determined.
    """
    return os.environ.get("KUBERNETES_NODE_NAME", "")


def get_digest() -> str:
    """Return the digest of the current Docker image.

    Returns
    -------
    str
        Digest of the Docker image this code is running inside, or the empty
        string if the digest could not be determined.
    """
    spec = os.environ.get("JUPYTER_IMAGE_SPEC", "")
    hash_marker = "@sha256:"
    hash_pos = spec.find(hash_marker)
    if hash_pos == -1:
        return ""
    return spec[hash_pos + len(hash_marker) :]


def get_access_token(
    tokenfile: str | Path | None = None, log: Any | None = None
) -> str:
    """Get the Gafaelfawr access token for the user.

    Determine the access token from the mounted location (nublado 3/2) or
    environment (any).  Prefer the mounted version since it can be updated,
    while the environment variable stays at whatever it was when the process
    was started.  Return the empty string if the token cannot be determined.
    """
    if tokenfile:
        return Path(tokenfile).read_text()
    base_dir = Path("/opt/lsst/software/jupyterlab")
    for candidate in (
        base_dir / "secrets" / "token",
        base_dir / "environment" / "ACCESS_TOKEN",
    ):
        with suppress(FileNotFoundError):
            return candidate.read_text()

    # If we got here, we couldn't find a file. Return the environment variable
    # if set, otherwise the empty string.
    return os.environ.get("ACCESS_TOKEN", "")



"""Tests for startup object."""

import configparser
import json
import os
from collections.abc import Iterable
from pathlib import Path

import pytest
import symbolicmode

import lsst.rsp
from lsst.rsp.startup.services.labrunner import LabRunner


@pytest.mark.usefixtures("_rsp_env")
def test_object() -> None:
    lr = LabRunner()
    assert lr._debug is False


@pytest.mark.usefixtures("_rsp_env")
def test_debug_object(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("DEBUG", "1")
    lr = LabRunner()
    assert lr._debug is True


#
# Environment methods
#


@pytest.mark.usefixtures("_rsp_env")
def test_cpu_vars(monkeypatch: pytest.MonkeyPatch) -> None:
    lr = LabRunner()
    lr._set_cpu_variables()
    assert lr._env["CPU_LIMIT"] == "1"
    # We need a new LabRunner each time, because it reads its environment
    # only at __init()__
    monkeypatch.setenv("CPU_LIMIT", "NaN")
    lr = LabRunner()
    lr._set_cpu_variables()
    assert lr._env["CPU_COUNT"] == "1"
    monkeypatch.setenv("CPU_LIMIT", "0.1")
    lr = LabRunner()
    lr._set_cpu_variables()
    assert lr._env["GOTO_NUM_THREADS"] == "1"
    monkeypatch.setenv("CPU_LIMIT", "3.1")
    lr = LabRunner()
    lr._set_cpu_variables()
    assert lr._env["MKL_DOMAIN_NUM_THREADS"] == "3"
    monkeypatch.setenv("CPU_LIMIT", "14")
    lr = LabRunner()
    lr._set_cpu_variables()
    assert lr._env["MPI_NUM_THREADS"] == "14"


# No test for set_image_digest() because we test that in utils_test


@pytest.mark.usefixtures("_rsp_env")
def test_expand_panda_tilde(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("PANDA_CONFIG_ROOT", "~")
    lr = LabRunner()
    lr._expand_panda_tilde()
    assert lr._env["PANDA_CONFIG_ROOT"] == os.environ["HOME"]
    monkeypatch.setenv("PANDA_CONFIG_ROOT", "~hambone")
    lr = LabRunner()
    lr._expand_panda_tilde()
    assert lr._env["PANDA_CONFIG_ROOT"] == os.environ["HOME"]
    monkeypatch.setenv("PANDA_CONFIG_ROOT", "~hambone/")
    lr = LabRunner()
    lr._expand_panda_tilde()
    assert lr._env["PANDA_CONFIG_ROOT"] == os.environ["HOME"]
    monkeypatch.setenv("PANDA_CONFIG_ROOT", "~whoopsi")
    lr = LabRunner()
    lr._expand_panda_tilde()
    assert lr._env["PANDA_CONFIG_ROOT"] == "~whoopsi"
    monkeypatch.setenv("PANDA_CONFIG_ROOT", "/etc/panda")
    lr = LabRunner()
    lr._expand_panda_tilde()
    assert lr._env["PANDA_CONFIG_ROOT"] == "/etc/panda"
    monkeypatch.setenv("PANDA_CONFIG_ROOT", "~/bar")
    lr = LabRunner()
    lr._expand_panda_tilde()
    assert lr._env["PANDA_CONFIG_ROOT"] == str(
        Path(os.environ["HOME"]) / "bar"
    )


@pytest.mark.usefixtures("_rsp_env")
def test_set_timeout_vars(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("NO_ACTIVITY_TIMEOUT", "300")
    lr = LabRunner()
    lr._set_timeout_variables()
    assert lr._env["NO_ACTIVITY_TIMEOUT"] == "300"
    assert "CULL_KERNEL_IDLE_TIMEOUT" not in lr._env


@pytest.mark.usefixtures("_rsp_env")
def test_set_launch_params(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("JUPYTERHUB_BASE_URL", "/nb/")
    monkeypatch.setenv("EXTERNAL_INSTANCE_URL", "https://lab.example.com:8443")
    lr = LabRunner()
    lr._set_launch_params()
    assert lr._stash["jupyterhub_path"] == "/nb/hub"
    assert lr._stash["external_host"] == "lab.example.com"


@pytest.mark.usefixtures("_rsp_env")
def test_set_firefly_variables(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("EXTERNAL_INSTANCE_URL", "https://lab.example.com:8443")
    lr = LabRunner()
    lr._set_firefly_variables()
    assert lr._env["FIREFLY_URL"] == "https://lab.example.com:8443/firefly"
    assert lr._env["FIREFLY_HTML"] == "slate.html"


@pytest.mark.usefixtures("_rsp_env")
def test_force_jupyter_prefer_env_path_false() -> None:
    lr = LabRunner()
    lr._force_jupyter_prefer_env_path_false()
    assert lr._env["JUPYTER_PREFER_ENV_PATH"] == "no"


@pytest.mark.usefixtures("_rsp_env")
def test_set_butler_credential_vars(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("AWS_SHARED_CREDENTIALS_FILE", "/etc/secret/aws.creds")
    monkeypatch.setenv("PGPASSFILE", "/etc/secret/pgpass")
    lr = LabRunner()
    lr._set_butler_credential_variables()
    assert lr._env["AWS_SHARED_CREDENTIALS_FILE"] == str(
        lr._home / ".lsst" / "aws.creds"
    )
    assert (
        lr._env["ORIG_AWS_SHARED_CREDENTIALS_FILE"] == "/etc/secret/aws.creds"
    )
    assert lr._env["PGPASSFILE"] == str(lr._home / ".lsst" / "pgpass")
    assert lr._env["ORIG_PGPASSFILE"] == "/etc/secret/pgpass"


#
# File manipulation tests
#


@pytest.mark.usefixtures("_rsp_env")
def test_copy_butler_credentials(monkeypatch: pytest.MonkeyPatch) -> None:
    td = lsst.rsp.startup.constants.TOP_DIR_PATH
    secret_dir = td / "jupyterlab" / "secrets"
    monkeypatch.setenv(
        "AWS_SHARED_CREDENTIALS_FILE", str(secret_dir / "aws-credentials.ini")
    )
    monkeypatch.setenv(
        "PGPASSFILE", str(secret_dir / "postgres-credentials.txt")
    )
    lr = LabRunner()
    pg = lr._home / ".lsst" / "postgres-credentials.txt"
    lines = pg.read_text().splitlines()
    aws = lr._home / ".lsst" / "aws-credentials.ini"
    for line in lines:
        if line.startswith("127.0.0.1:5432:db01:postgres:"):
            assert line.rsplit(":", maxsplit=1)[1] == "gets_overwritten"
        if line.startswith("127.0.0.1:5532:db02:postgres:"):
            assert line.rsplit(":", maxsplit=1)[1] == "should_stay"
    cp = configparser.ConfigParser()
    cp.read(str(aws))
    assert set(cp.sections()) == {"default", "tertiary"}
    assert cp["default"]["aws_secret_access_key"] == "gets_overwritten"
    assert cp["tertiary"]["aws_secret_access_key"] == "key03"
    lr._set_butler_credential_variables()
    lr._copy_butler_credentials()
    lines = pg.read_text().splitlines()
    aws = lr._home / ".lsst" / "aws-credentials.ini"
    for line in lines:
        if line.startswith("127.0.0.1:5432:db01:postgres:"):
            assert line.rsplit(":", maxsplit=1)[1] == "s33kr1t"
        if line.startswith("127.0.0.1:5532:db02:postgres:"):
            assert line.rsplit(":", maxsplit=1)[1] == "should_stay"
    cp = configparser.ConfigParser()
    cp.read(str(aws))
    assert set(cp.sections()) == {"default", "secondary", "tertiary"}
    assert cp["default"]["aws_secret_access_key"] == "key01"
    assert cp["secondary"]["aws_secret_access_key"] == "key02"
    assert cp["tertiary"]["aws_secret_access_key"] == "key03"


@pytest.mark.usefixtures("_rsp_env")
def test_copy_logging_profile(monkeypatch: pytest.MonkeyPatch) -> None:
    lr = LabRunner()
    pfile = (
        lr._home / ".ipython" / "profile_default" / "startup" / "20-logging.py"
    )
    td = lsst.rsp.startup.constants.TOP_DIR_PATH
    assert not pfile.exists()
    pfile.parent.mkdir(parents=True)
    lr._copy_logging_profile()
    assert pfile.exists()
    h_contents = pfile.read_text()
    sfile = td / "jupyterlab" / "20-logging.py"
    assert sfile.exists()
    s_contents = sfile.read_text()
    assert s_contents == h_contents
    h_contents += "\n# Locally modified\n"
    pfile.write_text(h_contents)
    lr._copy_logging_profile()
    new_contents = pfile.read_text()
    assert new_contents == h_contents
    assert new_contents != s_contents


@pytest.mark.usefixtures("_rsp_env")
def test_copy_dircolors(monkeypatch: pytest.MonkeyPatch) -> None:
    lr = LabRunner()
    assert not (lr._home / ".dir_colors").exists()
    lr._copy_dircolors()
    assert (lr._home / ".dir_colors").exists()


@pytest.mark.usefixtures("_rsp_env")
def test_copy_etc_skel(monkeypatch: pytest.MonkeyPatch) -> None:
    lr = LabRunner()
    assert not (lr._home / ".gitconfig").exists()
    assert not (lr._home / ".pythonrc").exists()
    etc = lsst.rsp.startup.constants.ETC_PATH
    prc = (etc / "skel" / ".pythonrc").read_text()
    prc += "\n# Local mods\n"
    (lr._home / ".pythonrc").write_text(prc)
    lr._copy_etc_skel()
    assert (lr._home / ".gitconfig").exists()
    sgc = (etc / "skel" / ".gitconfig").read_text()
    lgc = (lr._home / ".gitconfig").read_text()
    assert sgc == lgc
    src = (etc / "skel" / ".pythonrc").read_text()
    lrc = (lr._home / ".pythonrc").read_text()
    assert src != lrc
    assert (lr._home / "notebooks" / ".user_setups").exists()


#
# Git
#


@pytest.mark.usefixtures("_rsp_env")
def test_refresh_notebooks(
    monkeypatch: pytest.MonkeyPatch, git_repo: Path
) -> None:
    source_repo = git_repo
    monkeypatch.setenv("AUTO_REPO_SPECS", f"file://{source_repo!s}@main")
    lr = LabRunner()
    repo = lr._home / "notebooks" / source_repo.name
    assert not repo.exists()
    lr._refresh_notebooks()
    paths = (repo, repo / "README.md")
    assert _is_readonly(paths)
    lr._refresh_notebooks()
    assert _is_readonly(paths)
    for p in paths:
        symbolicmode.chmod(p, "u+w")
    assert not _is_readonly(paths)
    lr._refresh_notebooks()
    assert _is_readonly(paths)


def _is_readonly(paths: Iterable[Path]) -> bool:
    for p in paths:
        assert p.exists()
        mode = p.stat().st_mode
        mask = 0o222
        if mode & mask != 0:
            return False
    return True


@pytest.mark.usefixtures("_rsp_env")
def test_setup_gitlfs(monkeypatch: pytest.MonkeyPatch) -> None:
    lr = LabRunner()
    assert lr._check_for_git_lfs() is False
    lr._setup_gitlfs()
    assert lr._check_for_git_lfs() is True


#
# Interactive-mode-only tests
#


@pytest.mark.usefixtures("_rsp_env")
def test_increase_log_limit(monkeypatch: pytest.MonkeyPatch) -> None:
    lr = LabRunner()
    settings = (
        lr._home
        / ".jupyter"
        / "lab"
        / "user-settings"
        / "@jupyterlab"
        / "notebook-extension"
        / "tracker.jupyterlab.settings"
    )
    assert not settings.exists()
    lr._increase_log_limit()
    assert settings.exists()
    with settings.open() as f:
        obj = json.load(f)
    assert obj["maxNumberOutputs"] >= 10000


@pytest.mark.usefixtures("_rsp_env")
def test_manage_access_token(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("DEBUG", "1")
    token = "token-of-esteem"
    monkeypatch.setenv("ACCESS_TOKEN", token)
    td = lsst.rsp.startup.constants.TOP_DIR_PATH
    ctr_file = td / "jupyterlab" / "secrets" / "token"
    assert not ctr_file.exists()
    lr = LabRunner()
    tfile = lr._home / ".access_token"
    assert not tfile.exists()
    lr._manage_access_token()
    assert tfile.exists()
    assert tfile.read_text() == token
    tfile.unlink()
    ctr_file.write_text(token)
    assert ctr_file.exists()
    assert not tfile.exists()
    lr = LabRunner()
    lr._manage_access_token()
    assert tfile.exists()
    assert tfile.read_text() == token
    ctr_file.unlink()
    assert not ctr_file.exists()



"""Utility functions to get clients for TAP catalog search."""

import warnings

import pyvo
from deprecated import deprecated

from .utils import get_pyvo_auth, get_service_url


@deprecated(reason='Please use get_tap_service("tap")')
def get_catalog() -> pyvo.dal.TAPService:
    """Call ``get_tap_service("tap")`` (deprecated alias)."""
    return get_tap_service("tap")


@deprecated(reason='Please use get_tap_service("live")')
def get_obstap_service() -> pyvo.dal.TAPService:
    """Call ``get_tap_service("live")`` (deprecated alias)."""
    return get_tap_service("live")


def get_tap_service(*args: str) -> pyvo.dal.TAPService:
    """Construct a TAP service instance for the requested TAP service."""
    if len(args) == 0:
        warnings.warn(
            'get_tap_service() is deprecated, use get_tap_service("tap")',
            DeprecationWarning,
            stacklevel=2,
        )
        database = "tap"
    else:
        database = args[0]

    # We renamed the name of the TAP service from obstap
    # to live
    if database == "obstap":
        database = "live"

    if database in ("live", "tap", "ssotap"):
        tap_url = get_service_url(database)
    else:
        raise ValueError(f"{database} is not a valid tap service")

    # This is not ideal, but warning appears because require pyvo does
    # not register uws:Sync and uws:Async.  It's harmless.  The broadness
    # of the warning is unfortunately necessary since pyvo just uses
    # warnings.warn():
    #
    # https://github.com/astropy/pyvo/blob/
    # 81a50d7fd24428f17104a075bc0e1ac661ed6ea0/pyvo/utils/xml/elements.py#L418
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        return pyvo.dal.TAPService(tap_url, get_pyvo_auth())


def retrieve_query(query_url: str) -> pyvo.dal.AsyncTAPJob:
    """Retrieve job corresponding to a particular query URL."""
    # This is not ideal, but warning appears because require pyvo does
    # not register uws:Sync and uws:Async.  It's harmless.  The broadness
    # of the warning is unfortunately necessary since pyvo just uses
    # warnings.warn():
    #
    # https://github.com/astropy/pyvo/blob/
    # 81a50d7fd24428f17104a075bc0e1ac661ed6ea0/pyvo/utils/xml/elements.py#L418
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        return pyvo.dal.AsyncTAPJob(query_url, get_pyvo_auth())



"""RSP Lab launcher."""

import configparser
import contextlib
import datetime
import hashlib
import json
import os
import pwd
import shutil
import sys
from pathlib import Path
from subprocess import SubprocessError
from time import sleep
from typing import Any
from urllib.parse import urlparse

import structlog
import symbolicmode

from ... import get_access_token, get_digest
from ..constants import (
    APP_NAME,
    ETC_PATH,
    MAX_NUMBER_OUTPUTS,
    NONINTERACTIVE_CONFIG_PATH,
    PREVIOUS_LOGGING_CHECKSUMS,
    TOP_DIR_PATH,
)
from ..models.noninteractive import NonInteractiveExecutor
from ..storage.command import Command
from ..storage.logging import configure_logging

__all__ = ["LabRunner"]


class LabRunner:
    """Class to start JupyterLab using the environment supplied by
    JupyterHub and the Nublado controller.

    This environment is very Rubin-specific and opinionated, and will
    likely not work for anyone else's science platform.

    If that's you, use this for inspiration, but don't expect this to
    work out of the box.
    """

    def __init__(self) -> None:
        # We start with a copy of our own environment
        self._env = os.environ.copy()
        self._debug = bool(self._env.get("DEBUG", ""))
        configure_logging(debug=self._debug)
        self._logger = structlog.get_logger(APP_NAME)
        self._home = Path(self._env["HOME"])  # If unset, it's OK to die.
        if "JUPYTERHUB_BASE_URL" not in self._env:
            raise ValueError("'JUPYTERHUB_BASE_URL' must be set")
        self._user = ""
        self._stash: dict[str, str] = {}  # Used for settings we don't expose.
        self._cmd = Command(ignore_fail=True, logger=self._logger)

    def go(self) -> None:
        """Start the user lab."""
        # The LabRunner is not actually the first thing that launches when
        # we start a user lab.
        #
        # At the moment, the only Python 3 in the Lab container is the one that
        # is part of the DM stack conda environment.
        #
        # So to get far enough to even start the LabRunner, we need to have
        # already sourced the shell magic that sets up that conda env.
        #
        # However, even before we do that, we check the environment to see
        # whether it needs resetting.  That's actually kind of handy to do
        # before we start any Python process, because it's not impossible that
        # the user could have messed up their own Python environment so badly
        # that Python couldn't even get this far.
        #
        # When we split the stack Python from the Python-that-runs-JuptyerLab,
        # there will be no need to set up the stack before launching
        # Jupyterlab, and the likelihood that the user environment is so
        # corrupt that the user cannot get a terminal session open to purge
        # it themselves will lessen greatly.  At that point moving the
        # user-environment purge into lsst-rsp may be feasible.
        #

        # Set up environment variables that we'll need either to launch the
        # Lab or for the user's terminal environment
        self._configure_env()

        # Copy files into the user's home space.  If $HOME is not mounted
        # and writeable, things will go wrong here.
        self._copy_files_to_user_homedir()

        # Check out notebooks, and set up git-lfs
        self._setup_git()

        # Clear EUPS cache
        self._cmd.run("eups", "admin", "clearCache")

        # Decide between interactive and noninteractive start, do
        # things that change between those two, and launch the Lab
        self._launch()

    def _externalize(self, setting: str) -> str:
        # We build multiple settings by concatenating `EXTERNAL_INSTANCE_URL`
        # with some other string.  Make this robust by accepting either
        # a slash or no slash on the end of `EXTERNAL_INSTANCE_URL` and
        # returning a string with exactly one slash as a separator between
        # the external URL and the setting.
        ext_url = self._env["EXTERNAL_INSTANCE_URL"]
        return f"{ext_url.strip('/')}/{setting.lstrip('/')}"

    #
    # Next up, a big block of setting up our subprocess environment.
    #
    def _configure_env(self) -> None:
        self._logger.debug("Configuring environment for JupyterLab process")
        self._set_user()
        self._set_cpu_variables()
        self._set_image_digest()
        self._expand_panda_tilde()
        self._set_launch_params()
        self._set_firefly_variables()
        self._force_jupyter_prefer_env_path_false()
        self._set_butler_credential_variables()
        self._logger.debug("Lab process environment", env=self._env)

    def _set_user(self) -> None:
        self._logger.debug("Determining user name")
        user = self._env.get("USER", "")
        if not user:
            self._env["USER"] = pwd.getpwuid(os.getuid()).pw_name

    def _set_cpu_variables(self) -> None:
        self._logger.debug("Setting CPU threading variables")
        try:
            cpu_limit = int(float(self._env.get("CPU_LIMIT", "1")))
        except ValueError:
            cpu_limit = 1
        if cpu_limit < 1:
            cpu_limit = 1
        cpu_limit_str = str(cpu_limit)
        for vname in (
            "CPU_LIMIT",
            "CPU_COUNT",
            "GOTO_NUM_THREADS",
            "MKL_DOMAIN_NUM_THREADS",
            "MPI_NUM_THREADS",
            "NUMEXPR_NUM_THREADS",
            "NUMEXPR_MAX_THREADS",
            "OMP_NUM_THREADS",
            "OPENBLAS_NUM_THREADS",
            "RAYON_NUM_THREADS",
        ):
            self._env[vname] = cpu_limit_str
            self._logger.debug(f"Set '{vname}' -> '{cpu_limit_str}'")

    def _set_image_digest(self) -> None:
        self._logger.debug("Setting image digest if available")
        # get_digest() is already a helper function in our parent package.
        digest = get_digest()
        if digest:
            self._logger.debug(f"Set image digest to '{digest}'")
            self._env["IMAGE_DIGEST"] = digest
        else:
            self._logger.debug("Could not get image digest")

    def _expand_panda_tilde(self) -> None:
        self._logger.debug("Expanding tilde in PANDA_CONFIG_ROOT, if needed")
        if "PANDA_CONFIG_ROOT" in self._env:
            username = self._env["USER"]
            path = Path(self._env["PANDA_CONFIG_ROOT"])
            path_parts = path.parts
            if path_parts[0] in ("~", f"~{username}"):
                new_path = Path(self._home, *path_parts[1:])
                self._logger.debug(
                    f"Replacing PANDA_CONFIG_ROOT '{path!s}'"
                    f"with '{new_path!s}'"
                )
                self._env["PANDA_CONFIG_ROOT"] = str(new_path)
            elif path_parts[0].startswith("~"):
                self._logger.warning(f"Cannot expand tilde in '{path!s}'")

    def _set_launch_params(self) -> None:
        # We're getting rid of the complicated stuff based on
        # HUB_SERVICE_HOST, since that was pre-version-3 nublado.
        self._logger.debug("Setting launch parameters")
        base_url = self._env["JUPYTERHUB_BASE_URL"]
        jh_path = f"{base_url}hub"
        ext_url = self._env.get("EXTERNAL_INSTANCE_URL", "")
        host = urlparse(ext_url).hostname or ""

        self._stash["jupyterhub_path"] = jh_path
        self._stash["external_host"] = host
        self._logger.debug(
            f"Set host to '{host}', and Hub path to '{jh_path}'"
        )

    def _set_firefly_variables(self) -> None:
        self._logger.debug("Setting firefly variables")
        firefly_route = self._env.get("FIREFLY_ROUTE", "/firefly")
        self._env["FIREFLY_URL"] = self._externalize(firefly_route)
        self._logger.debug(f"Firefly URL -> '{self._env['FIREFLY_URL']}'")
        # This determines the landing page on the Firefly service.  For
        # the RSP, it's always `slate.html`.
        self._env["FIREFLY_HTML"] = "slate.html"

    def _force_jupyter_prefer_env_path_false(self) -> None:
        # cf https://discourse.jupyter.org/t/jupyter-paths-priority-order/7771
        # and https://jupyter-core.readthedocs.io/en/latest/changelog.html#id63
        #
        # As long as we're running from the stack Python, we need to ensure
        # this is turned off.
        self._logger.debug("Forcing JUPYTER_PREFER_ENV_PATH to 'no'")
        self._env["JUPYTER_PREFER_ENV_PATH"] = "no"

    def _set_butler_credential_variables(self) -> None:
        # We split this up into environment manipulation and later
        # file substitution.  This is the environment part.
        self._logger.debug("Setting Butler credential variables")
        cred_dir = self._home / ".lsst"
        if "AWS_SHARED_CREDENTIALS_FILE" in self._env:
            awsname = Path(self._env["AWS_SHARED_CREDENTIALS_FILE"]).name
            self._env["ORIG_AWS_SHARED_CREDENTIALS_FILE"] = self._env[
                "AWS_SHARED_CREDENTIALS_FILE"
            ]
            newaws = str(cred_dir / awsname)
            self._env["AWS_SHARED_CREDENTIALS_FILE"] = newaws
            self._logger.debug(
                f"Set 'AWS_SHARED_CREDENTIALS_FILE' -> '{newaws}'"
            )
        if "PGPASSFILE" in self._env:
            pgpname = Path(self._env["PGPASSFILE"]).name
            newpg = str(cred_dir / pgpname)
            self._env["ORIG_PGPASSFILE"] = self._env["PGPASSFILE"]
            self._env["PGPASSFILE"] = newpg
            self._logger.debug(f"Set 'PGPASSFILE' -> '{newpg}'")

    #
    # The second big block is a bunch of file manipulation.
    #
    def _copy_files_to_user_homedir(self) -> None:
        self._logger.debug("Copying files to user home directory")
        self._copy_butler_credentials()
        self._copy_logging_profile()
        self._copy_dircolors()
        self._copy_etc_skel()

    def _copy_butler_credentials(self) -> None:
        if "AWS_SHARED_CREDENTIALS_FILE" in self._env:
            self._merge_aws_creds()
        if "PGPASSFILE" in self._env:
            self._merge_pgpass()

    def _merge_aws_creds(self) -> None:
        #
        # Merge the config in the original credentials file and the one
        # in our homedir.  For any given section, we assume that the
        # information in the container ("original credentials files")
        # is correct, but leave any other user config alone.
        #
        hc_path = Path(self._env["AWS_SHARED_CREDENTIALS_FILE"])
        hc_path.touch(mode=0o600, exist_ok=True)
        home_config = configparser.ConfigParser()
        home_config.read(str(hc_path))
        ro_config = configparser.ConfigParser()
        ro_config.read(self._env["ORIG_AWS_SHARED_CREDENTIALS_FILE"])
        for sect in ro_config.sections():
            home_config[sect] = ro_config[sect]
        with hc_path.open("w") as f:
            home_config.write(f)

    def _merge_pgpass(self) -> None:
        #
        # Same as above, but for pgpass files.
        #
        config = {}
        # Get current config from homedir
        home_pgpass = Path(self._env["PGPASSFILE"])
        home_pgpass.touch(mode=0o600, exist_ok=True)
        lines = home_pgpass.read_text().splitlines()
        for line in lines:
            if ":" not in line:
                continue
            connection, passwd = line.rsplit(":", maxsplit=1)
            config[connection] = passwd.rstrip()
        # Update config from container-supplied one
        ro_pgpass = Path(self._env["ORIG_PGPASSFILE"])
        lines = ro_pgpass.read_text().splitlines()
        for line in lines:
            if ":" not in line:
                continue
            connection, passwd = line.rsplit(":", maxsplit=1)
            config[connection] = passwd.rstrip()
        with home_pgpass.open("w") as f:
            for connection in config:
                f.write(f"{connection}:{config[connection]}\n")

    def _copy_logging_profile(self) -> None:
        self._logger.debug("Copying logging profile if needed")
        user_profile = (
            self._home
            / ".ipython"
            / "profile_default"
            / "startup"
            / "20-logging.py"
        )
        #
        # We have a list of previous supplied versions of 20-logging.py.
        #
        # If the one we have has a hash that matches any of those, then
        # there is a new container-supplied 20-logging.py that should replace
        # it.  However, if we have a 20-logging.py that does not match
        # any of those, then it has been locally modified, and we should
        # not replace it.  If we don't have one at all, we need to copy it
        # into place.
        #
        copy = False
        if not user_profile.is_file():
            copy = True  # It doesn't exist, so we need one.
        else:
            user_loghash = hashlib.sha256(
                user_profile.read_bytes()
            ).hexdigest()
            if user_loghash in PREVIOUS_LOGGING_CHECKSUMS:
                self._logger.debug(
                    f"User log profile '{user_loghash}' is"
                    " out-of-date; replacing with current version."
                )
                copy = True
        if copy:
            user_profile.write_bytes(
                (TOP_DIR_PATH / "jupyterlab" / "20-logging.py").read_bytes()
            )

    def _copy_dircolors(self) -> None:
        self._logger.debug("Copying dircolors if needed")
        if not (self._home / ".dir_colors").exists():
            self._logger.debug("Copying dircolors")
            dc = ETC_PATH / "dircolors.ansi-universal"
            dc_txt = dc.read_text()
            (self._home / ".dir_colors").write_text(dc_txt)
        else:
            self._logger.debug("Copying dircolors not needed")

    def _copy_etc_skel(self) -> None:
        self._logger.debug("Copying files from /etc/skel if they don't exist")
        etc_skel = ETC_PATH / "skel"
        # alas, Path.walk() requires Python 3.12, which isn't in the
        # stack containers yet.
        contents = os.walk(etc_skel)
        #
        # We assume that if the file exists at all, we should leave it alone.
        # Users are allowed to modify these, after all.
        #
        for entry in contents:
            dirs = [Path(x) for x in entry[1]]
            files = [Path(x) for x in entry[2]]
            # Determine what the destination directory should be
            if entry[0] == str(etc_skel):
                current_dir = self._home
            else:
                current_dir = self._home / entry[0][(len(str(etc_skel)) + 1) :]
            # For each directory in the tree at this level:
            # if we don't already have one in our directory, make it.
            for d_item in dirs:
                if not (current_dir / d_item).is_dir():
                    (current_dir / d_item).mkdir()
                    self._logger.debug(f"Creating {current_dir / d_item!s}")
            # For each file in the tree at this level:
            # if we don't already have one in our directory, copy the
            # contents.
            for f_item in files:
                if not (current_dir / f_item).exists():
                    src = Path(entry[0] / f_item)
                    self._logger.debug(f"Creating {current_dir / f_item!s}")
                    src_contents = src.read_bytes()
                    (current_dir / f_item).write_bytes(src_contents)

    def _setup_git(self) -> None:
        # Refresh standard notebooks
        self._refresh_notebooks()
        # Set up Git LFS
        self._setup_gitlfs()

    def _refresh_notebooks(self) -> None:
        # Find the notebook specs.  I think we can ditch our fallbacks now.
        self._logger.debug("Refreshing notebooks")
        urls = self._env.get("AUTO_REPO_SPECS", "").split(",")
        if not urls:
            self._logger.debug("No repos listed in 'AUTO_REPO_SPECS'")
            return
        timeout = 30  # Probably don't need to parameterize it.
        reloc_msg = ""
        now = datetime.datetime.now(datetime.UTC).isoformat()
        for url in urls:
            try:
                repo, branch = url.split("@", maxsplit=1)
            except ValueError:
                self._logger.warning(
                    "Could not get repo/branch information from"
                    f" '{self._env['AUTO_REPO_SPECS']}'"
                )
                return
            repo_path = urlparse(repo).path
            repo_name = Path(repo_path).name
            if repo_name.endswith(".git"):
                repo_name = repo_name[:-4]
            dirname = self._home / "notebooks" / repo_name
            if dirname.is_dir():
                # We're going to make the simplifying assumption that the
                # user owns the directory and is in the directory's group.
                # If not, probably a lot else has already gone wrong.
                can_write = dirname.stat().st_mode & 0o222
                if can_write:
                    self._logger.debug(f"'{dirname!s}' is writeable; moving")
                    newname = Path(f"{dirname!s}.{now}")
                    reloc_msg += f"* '{dirname!s}' -> '{newname!s}'\n"
                    # We're also going to assume the user DOES have write
                    # permission in the parent directory.  Again, if not,
                    # terrible things probably already happened.
                    dirname.rename(newname)
                else:
                    # If the repository exists and is not writeable, and has
                    # the same last commit as the remote, then we don't
                    # need to update it.
                    if self._compare_local_and_remote(
                        dirname, branch, timeout
                    ):
                        self._logger.debug(f"'{dirname!s}' is r/o and current")
                        continue  # Up-to-date; we don't need to do anything
                    # It's writeable or stale; re-clone.
                    self._logger.debug(f"Need to remove '{dirname!s}'")
                    symbolicmode.chmod(dirname, "u+w", recurse=True)
                    shutil.rmtree(dirname)
            # If the directory existed, it's gone now.
            self._logger.debug(f"Cloning {repo}@{branch}")
            proc = self._cmd.run(
                "git",
                "clone",
                "--depth",
                "1",
                repo,
                "-b",
                branch,
                str(dirname),
                timeout=timeout,
            )
            if proc.returncode != 0:
                self._logger.error("git clone failed", proc=proc)
                return
            symbolicmode.chmod(dirname, "a-w", recurse=True)
        if reloc_msg:
            hdr = (
                "# Directory relocation\n\n"
                "The following directories were writeable, and were moved:\n"
                "\n"
                "\n"
            )
            reloc_msg = hdr + reloc_msg
            (self._home / "notebooks" / "00_README_RELOCATION.md").write_text(
                reloc_msg
            )

        self._logger.debug("Refreshed notebooks")

    def _compare_local_and_remote(
        self, path: Path, branch: str, timeout: int
    ) -> bool:
        # Returns True if git repo checked out to path has the same
        # commit hash as the remote.
        #
        # Git wants you to be in the working tree
        rx = self._cmd.run(
            "git",
            "rev-parse",
            "HEAD",
            cwd=path,
            timeout=timeout,
        )
        local_sha = rx.stdout.strip() if rx else None
        if not local_sha:
            self._logger.error(f"Could not determine local SHA for '{path!s}'")
            return False
        rx = self._cmd.run(
            "git",
            "config",
            "--get",
            "remote.origin.url",
            timeout=timeout,
        )
        remote = rx.stdout.strip() if rx else None
        if not remote:
            self._logger.error(
                "Could not determine git remote origin for" f"'{path!s}'"
            )
            return False
        rx = self._cmd.run(
            "git",
            "ls-remote",
            remote,
            timeout=timeout,
        )
        ls_remote = rx.stdout.strip() if rx else None
        lsr_lines = ls_remote.split("\n")
        remote_sha = None
        for line in lsr_lines:
            line.strip()
            if line.endswith(f"\trefs/heads/{branch}"):
                remote_sha = line.split()[0]
                break
        if not remote_sha:
            self._logger.error("Could not determine SHA for {remote}")
            return False
        self._logger.debug(f"local /remote SHA: {local_sha}/{remote_sha}")
        return local_sha == remote_sha

    def _setup_gitlfs(self) -> None:
        # Check for git-lfs
        self._logger.debug("Installing Git LFS if needed")
        if not self._check_for_git_lfs():
            self._cmd.run("git", "lfs", "install")
            self._logger.debug("Git LFS installed")

    def _check_for_git_lfs(self) -> bool:
        gitconfig = self._home / ".gitconfig"
        if gitconfig.is_file():
            gc = gitconfig.read_text().splitlines()
            for line in gc:
                line.strip()
                if line == '[filter "lfs"]':
                    return True
        return False

    def _launch(self) -> None:
        if bool(self._env.get("NONINTERACTIVE", "")):
            self._start_noninteractive()
            # We exec a lab; control never returns here
        self._modify_interactive_settings()
        self._start()

    def _modify_interactive_settings(self) -> None:
        self._logger.debug("Modifying interactive settings if needed")
        self._manage_access_token()
        self._increase_log_limit()

    def _increase_log_limit(self) -> None:
        self._logger.debug("Increasing log limit if needed")
        settings: dict[str, Any] = {}
        settings_dir = (
            self._home
            / ".jupyter"
            / "lab"
            / "user-settings"
            / "@jupyterlab"
            / "notebook-extension"
        )
        settings_dir.mkdir(parents=True, exist_ok=True)
        settings_file = settings_dir / "tracker.jupyterlab.settings"
        if settings_file.is_file():
            with settings_file.open() as f:
                settings = json.load(f)
        current_limit = settings.get("maxNumberOutputs", 0)
        if current_limit < MAX_NUMBER_OUTPUTS:
            self._logger.warning(
                f"Changing maxNumberOutputs in {settings_file!s}"
                f" from {current_limit} to {MAX_NUMBER_OUTPUTS}"
            )
            settings["maxNumberOutputs"] = MAX_NUMBER_OUTPUTS
            with settings_file.open("w") as f:
                json.dump(settings, f, sort_keys=True, indent=4)
        else:
            self._logger.debug("Log limit increase not needed")

    def _manage_access_token(self) -> None:
        self._logger.debug("Updating access token")
        tokfile = self._home / ".access_token"
        tokfile.unlink(missing_ok=True)
        ctr_token = (
            TOP_DIR_PATH / "software" / "jupyterlab" / "secrets" / "token"
        )
        if ctr_token.exists():
            self._logger.debug(f"Symlinking {tokfile!s}->{ctr_token!s}")
            tokfile.symlink_to(ctr_token)
            with contextlib.suppress(NotImplementedError):
                tokfile.chmod(0o600, follow_symlinks=False)
            return
        self._logger.debug("Did not find container token file")
        token = get_access_token()
        if token:
            tokfile.touch(mode=0o600)
            tokfile.write_text(token)
            self._logger.debug(f"Created {tokfile}")
        else:
            self._logger.debug("Could not determine access token")

    def _start_noninteractive(self) -> None:
        launcher = NonInteractiveExecutor.from_config(
            NONINTERACTIVE_CONFIG_PATH
        )
        launcher.execute(env=self._env)

    def _set_timeout_variables(self) -> list[str]:
        timeout_map = {
            "NO_ACTIVITY_TIMEOUT": "ServerApp.shutdown_no_activity_timeout",
            "CULL_KERNEL_IDLE_TIMEOUT": (
                "MappingKernelManager.cull_idle_timeout"
            ),
            "CULL_KERNEL_CONNECTED": "MappingKernelManager.cull_connected",
            "CULL_KERNEL_INTERVAL": "MappingKernelManager.cull_interval",
            "CULL_TERMINAL_INACTIVE_TIMEOUT": (
                "TerminalManager.cull_inactive_timeout"
            ),
            "CULL_TERMINAL_INTERVAL": "TerminalManager.cull_interval",
        }
        result: list[str] = []
        for setting in timeout_map:
            val = self._env.get(setting, "")
            if val:
                result.append(f"--{timeout_map[setting]}={val}")
        return result

    def _start(self) -> None:
        log_level = "DEBUG" if self._debug else "INFO"
        cmd = [
            "python3",
            "-s",
            "-m",
            "jupyter",
            "labhub",
            "--ip=0.0.0.0",
            "--port=8888",
            "--no-browser",
            f"--notebook-dir={self._home!s}",
            f"--hub-prefix={self._stash['jupyterhub_path']}",
            f"--hub-host={self._stash['external_host']}",
            f"--log-level={log_level}",
            "--ContentsManager.allow_hidden=True",
            "--FileContentsManager.hide_globs=[]",
            "--KernelSpecManager.ensure_native_kernel=False",
            "--QtExporter.enabled=False",
            "--PDFExporter.enabled=False",
            "--WebPDFExporter.allow_chromium_download=True",
            "--MappingKernelManager.default_kernel_name=lsst",
            "--LabApp.check_for_updates_class=jupyterlab.NeverCheckForUpdate",
        ]
        cmd.extend(self._set_timeout_variables())
        self._logger.debug("Command to run:", command=cmd)
        if self._debug:
            # Maybe we want to parameterize these?
            retries = 10
            sleep_interval = 60
            for i in range(retries):
                self._logger.debug(f"Lab spawn attempt {i+1}/{retries}:")
                try:
                    proc = self._cmd.run(*cmd, env=self._env)
                except SubprocessError as exc:
                    self._logger.exception(
                        f"Command {cmd} failed to run", exc=exc
                    )
                if proc:
                    if proc.returncode:
                        self._logger.error(
                            f"Lab exited with returncode {proc.returncode}",
                            proc=proc,
                        )
                    else:
                        self._logger.warning(
                            "Lab process exited with returncode 0", proc=proc
                        )
                else:
                    self._logger.error(f"Lab process {cmd} failed to run")
                self._logger.info(f"Waiting for {sleep_interval}s")
                sleep(sleep_interval)
            self._logger.debug("Exiting")
            sys.exit(0)
        # Flush open files before exec()
        sys.stdout.flush()
        sys.stderr.flush()
        # In non-debug mode, we don't use a subprocess: we exec the
        # Jupyter Python process directly.  We use os.execvpe() because we
        # want the Python in the path (which we currently know to be the
        # stack Python), we have a list of arguments we just created, and
        # we want to pass the environment we built up.
        os.execvpe(cmd[0], cmd, env=self._env)



"""Tests for utility functions."""

from __future__ import annotations

import pytest

from lsst.rsp import format_bytes
from lsst.rsp.utils import get_digest, get_service_url


def test_format_bytes() -> None:
    """Test human-readable names for numeric byte inputs."""
    assert format_bytes(1) == "1 B"
    assert format_bytes(1234) == "1.23 kB"
    assert format_bytes(12345678) == "12.35 MB"
    assert format_bytes(1234567890) == "1.23 GB"
    assert format_bytes(1234567890000) == "1.23 TB"
    assert format_bytes(1234567890000000) == "1.23 PB"


def test_get_digest(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("JUPYTER_IMAGE_SPEC", "sciplat-lab@sha256:abcde")
    digest = get_digest()
    assert digest == "abcde"


def test_get_digest_fail(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("JUPYTER_IMAGE_SPEC", "sciplat-lab:w_2024_01")
    digest = get_digest()
    assert digest == ""


def test_get_service_url(monkeypatch: pytest.MonkeyPatch) -> None:
    """Ensure there are no doubled slashes."""
    monkeypatch.setenv("EXTERNAL_INSTANCE_URL", "https://test.example.com/")
    monkeypatch.setenv("TAP_ROUTE", "/api/tap")
    assert get_service_url("tap") == "https://test.example.com/api/tap"
